use std::path::Path;

use monoio::fs;
use ocl::{builders::BufferBuilder, Device, Program, Queue};
use serde::{Deserialize, Serialize};

use crate::{layer::Layer, Error};

/// index of precompiled opencl program
pub const PROGRAM_FORWARD: usize = 0;

/// index of precompiled opencl program
pub const PROGRAM_BACKWARD: usize = 1;

pub const LEARN_RATE: f32 = 0.05;

/// AppState
pub struct Context {
    pub context: ocl::Context,
    pub device: Device,
    pub queue: Queue,
    pub program: [Program; 2],
}

impl Context {
    pub fn new() -> Self {
        let context = ocl::Context::builder()
            .devices(Device::specifier().first())
            .build()
            .unwrap();
        let device = context.devices()[0];
        let queue = Queue::new(&context, device, None).unwrap();

        let forward = Program::builder()
            .src(include_str!("../cl/forward.cl"))
            .devices(device)
            .build(&context)
            .unwrap();
        let backward = Program::builder()
            .src(include_str!("../cl/backward.cl"))
            .devices(device)
            .build(&context)
            .unwrap();

        Context {
            context,
            device,
            queue,
            program: [forward, backward],
        }
    }
}

#[derive(Serialize, Deserialize)]
struct LayersDump(Vec<u8>);

pub struct Layers(Vec<Layer>);

impl Layers {
    /// load from file
    pub async fn load_from(path: impl AsRef<Path>) -> Result<Self, Error> {
        let file = fs::File::open(path.as_ref()).await?;

        todo!()
    }
    /// create layer with random param
    pub fn new() -> Self {
        todo!()
    }
    /// train
    pub fn train(&mut self, ctx: &Context, data: Vec<f32>) -> Result<(), Error> {
        let input = unsafe {
            BufferBuilder::new()
                .context(&ctx.context)
                .use_host_slice(&data)
                .len(data.len())
                .build()?
        };
        let mut input = &input;

        for layer in &mut self.0 {
            layer.forward(ctx, input)?;
            input = &layer.activate;
        }

        let mut delta = BufferBuilder::<f32>::new()
            .context(&ctx.context)
            .len(1024)
            .fill_val(1.0)
            .build()?;

        for layer in self.0.iter_mut().rev() {
            layer.backward(ctx, &mut delta)?;
        }

        drop(data);
        
        Ok(())
    }
}
